@@ -1,4 +1,4 @@
-/* $OpenBSD: kex.c,v 1.79 2007/06/05 06:52:37 djm Exp $ */
+/* $OpenBSD: kex.c,v 1.80 2008/09/06 12:24:13 djm Exp $ */
 /*
  * Copyright (c) 2000, 2001 Markus Friedl.  All rights reserved.
  *
@@ -48,8 +48,6 @@
 
 #define KEX_COOKIE_LEN	16
 
-extern const EVP_MD *evp_ssh_sha256(void);
-
 /* prototype */
 static void kex_kexinit_finish(Kex *);
 static void kex_choose_conf(Kex *);
@@ -315,7 +313,7 @@ choose_kex(Kex *k, char *client, char *server)
 		k->evp_md = EVP_sha1();
 	} else if (strcmp(k->name, KEX_DHGEX_SHA256) == 0) {
 		k->kex_type = KEX_DH_GEX_SHA256;
-		k->evp_md = evp_ssh_sha256();
+		k->evp_md = EVP_sha256();
 	} else
 		fatal("bad kex alg %s", k->name);
 }
